#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
from .catom import Member, DefaultValue, Validate


class Coerced(Member):
    """ A member which will coerce a value to a given instance type.

    Unlike Typed or Instance, a Coerced value is not intended to be
    set to None.

    """
    __slots__ = ()

    def __init__(self, kind, args=None, kwargs=None, factory=None,
                 coercer=None):
        """ Initialize a Coerced.

        Parameters
        ----------
        kind : type or tuple of types
            The allowable types for the value.

        args : tuple, optional
            If 'factory' is None, then 'kind' is a callable type and
            these arguments will be passed to the constructor to create
            the default value.

        kwargs : dict, optional
            If 'factory' is None, then 'kind' is a callable type and
            these keywords will be passed to the constructor to create
            the default value.

        factory : callable, optional
            An optional callable which takes no arguments and returns
            the default value for the member. If this is not provided
            then 'args' and 'kwargs' should be provided, as 'kind' will
            be used to generate the default value.

        coercer : callable, optional
            An optional callable which takes the value and returns the
            coerced value. If this is not given, then 'kind' must be a
            callable type which will be called with the value to coerce
            the value to the appropriate type.

        """
        if factory is not None:
            self.set_default_value_mode(DefaultValue.CallObject, factory)
        else:
            args = args or ()
            kwargs = kwargs or {}
            factory = lambda: kind(*args, **kwargs)
            self.set_default_value_mode(DefaultValue.CallObject, factory)
        self.set_validate_mode(Validate.Coerced, (kind, coercer or kind))



#------------------------------------------------------------------------------
# Copyright (c) 2013-2018, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
import sys

IS_PY3 = sys.version_info >= (3,)


# From future.utils
def with_metaclass(meta, *bases):
    """
    Function from jinja2/_compat.py. License: BSD.

    Use it like this::

        class BaseForm(object):
            pass

        class FormType(type):
            pass

        class Form(with_metaclass(FormType, BaseForm)):
            pass

    This requires a bit of explanation: the basic idea is to make a
    dummy metaclass for one level of class instantiation that replaces
    itself with the actual metaclass.  Because of internal type checks
    we also need to make sure that we downgrade the custom metaclass
    for one level to something closer to type (that's why __call__ and
    __init__ comes back from type etc.).

    This has the advantage over six.with_metaclass of not introducing
    dummy classes into the final MRO.
    """
    class MetaClass(meta):
        __call__ = type.__call__
        __init__ = type.__init__

        def __new__(cls, name, this_bases, d):
            if this_bases is None:
                return type.__new__(cls, name, (), d)
            return meta(name, bases, d)
    return MetaClass('temporary_class', None, {})


if IS_PY3:
    import builtins
    bytes = builtins.bytes
    int = builtins.int
    str = builtins.str
    long = int

    # Alternatively this could be used
    # From past.types.basestring
    class BaseBaseString(type):
        def __instancecheck__(cls, instance):
            return isinstance(instance, (bytes, str))

    def __subclasshook__(cls, thing):
        raise NotImplemented  # pragma: no cover

    class basestring(with_metaclass(BaseBaseString)):
        """
        A minimal backport of the Python 2 basestring type to Py3
        """


else:
    from __builtin__ import bytes, basestring, int, long
    str = unicode

__all__ = ('IS_PY3', 'with_metaclass', 'int', 'long', 'bytes', 'str',
           'basestring')



#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
from .catom import Member, Validate, GetAttr, SetAttr, DelAttr


class Event(Member):
    """ A member which acts like a stateless event.

    """
    __slots__ = ()

    def __init__(self, kind=None):
        """ Initialize an Event.

        Parameters
        ----------
        kind : type or Member, optional
            The type of argument which may be emitted by the event or
            a Member which will validate the argument which can be
            emitted. The default is None and indicates no validation
            will be performed.

        """
        self.set_getattr_mode(GetAttr.Event, None)
        self.set_setattr_mode(SetAttr.Event, None)
        self.set_delattr_mode(DelAttr.Event, None)
        if kind is not None:
            if isinstance(kind, Member):
                self.set_validate_mode(Validate.Delegate, kind)
            else:
                self.set_validate_mode(Validate.Instance, kind)

    def set_name(self, name):
        """ A reimplemented parent class method.

        This method ensures that the delegate name is also set, if a
        delegate validator is being used.

        """
        super(Event, self).set_name(name)
        mode, kind = self.validate_mode
        if isinstance(kind, Member):
            kind.set_name(name)

    def set_index(self, index):
        """ A reimplemented parent class method.

        This method ensures that the delegate index is also set, if a
        delegate validator is being used.

        """
        super(Event, self).set_index(index)
        mode, kind = self.validate_mode
        if isinstance(kind, Member):
            kind.set_index(index)
